/**
 *   '$Author: bojilova $'
 *     '$Date: 2001/08/08 19:56:35 $'
 * '$Revision: 1.17 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
package edu.ucsb.nceas.metacat;

import java.sql.*;
import java.util.*;
import java.net.*;

public class RelationHandler //implements Runnable
{
  private Thread btThread = null;
  private Connection conn = null;
  private String docid = null;
  MetaCatUtil util = new MetaCatUtil();
  
  /** 
   * Constructor for this class.  finds all of the relations to a single xml
   * document and writes them to the database.  This includes transitive
   * relations.
   * @param docid the ID of the XML document to index.
   */
  public RelationHandler(String docid, Connection conn)
              throws McdbException, SQLException, AccessionNumberException
  {
    this.conn = conn;
    this.docid = docid;
    putRelations();
  }
  
  /**
   * insert the relations specified in the triples into xml_relation table
   */ 
  private void putRelations() 
              throws McdbException, SQLException, AccessionNumberException
  {
    String packagetype = null;
    String subject = null;
    String subDoctype = null;
    String relationship = null;
    String object = null;
    String objDoctype = null;
    PreparedStatement pstmt = null; // to get the relations from xml_nodes
    PreparedStatement tstmt = null; // to insert each relation into xml_relation
    MetaCatUtil.debugMessage("Running relation handler!");

    /* 
     * PSEUDO-CODE ALGORITHM:
     * deletes all of the relations with a docid of @docid.
     * for each new relation r in xml_nodes
     *   put r into xml_relation
     */

    //DocumentImpl xmldoc = new DocumentImpl(conn,docid,false);
    packagetype = (new DocumentImpl(conn,docid,false)).getDoctype();

    // first delete the relations for this package document if any
    deleteRelations(docid);

    // to put the new relations get them out of xml_nodes
    pstmt = conn.prepareStatement(QuerySpecification.printPackageSQL(docid));
    pstmt.execute();
    ResultSet rs = pstmt.getResultSet();
    boolean hasmorerows = rs.next();
    if (hasmorerows) {
      tstmt = conn.prepareStatement("INSERT INTO xml_relation (" +
                                    "docid,packagetype,subject,subdoctype," +
                                    "relationship, object, objdoctype) " + 
                                    "VALUES (?, ?, ?, ?, ?, ?, ?)");
    }
    while(hasmorerows) {
      subject = rs.getString(1);
      relationship = rs.getString(2);
      object = rs.getString(3);

      // cut out the revision number for subject and object
      subject = (new DocumentIdentifier(subject)).getIdentifier();
      object = (new DocumentIdentifier(object)).getIdentifier();

      //subDoctype and objDoctype are N/A
      subDoctype = null;
      objDoctype = null;

      //put the new relation into xml_relation
      tstmt.setString(1, docid);
      tstmt.setString(2, packagetype);
      tstmt.setString(3, subject);
      tstmt.setString(4, subDoctype);
      tstmt.setString(5, relationship);
      tstmt.setString(6, object);
      tstmt.setString(7, objDoctype);
      tstmt.execute(); 
        
      hasmorerows = rs.next();
    }
      
    if ( tstmt != null ) {
      tstmt.close();
    }
    pstmt.close();
    conn.commit();
  }
  
  /**
   * Deletes all of the relations with a docid of 'docid'.
   * @param docid the docid of the package which relations to delete.
   */
  public void deleteRelations(String docid) throws SQLException
  {
    try {
      PreparedStatement pstmt = conn.prepareStatement(
                                "DELETE FROM xml_relation " +
                                "WHERE docid = '" + docid + "'");
      pstmt.execute();
      pstmt.close();
    } catch(SQLException e) {
      MetaCatUtil.debugMessage("error in RelationHandler.deleteRelations(): " + 
                          e.getMessage());
      System.out.println("error in RelationHandler.deleteRelations(): " + 
                          e.getMessage());
      throw e;
    }
  }

  /**
   * Get the access file id for a package
   * @param docid the document identifier of the package
   * @return the document identifier of the access file for that package
   */
  public String getAccessFileID(String docid) throws SQLException
  {
    String aclid = null;
    PreparedStatement pstmt = 
      conn.prepareStatement("SELECT docid FROM xml_documents " +
                            "WHERE docid in " +
                            "(SELECT subject FROM xml_relation " +
                            "WHERE docid = ?) " +
                            "AND doctype = ?");
    pstmt.setString(1, docid);
    pstmt.setString(2, MetaCatUtil.getOption("accessdoctype"));
    pstmt.execute();
    ResultSet rs = pstmt.getResultSet();
    boolean hasRow = rs.next();
    if (hasRow) {
      aclid = rs.getString(1);
    }
    pstmt.close();
    
    return aclid;
  }

}
